package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  
import javax.swing.*;
import java.io.*;
import java.awt.image.*;

class LeakedLight extends JFrame{ 
	int raswidth, rasheight;
	int secnum;
	int npix=1;
	int rgbchan=0;
	int numDeconvolutionIterations=0;
	float [] leaksum,origSect;
	double SubfacForSubtractionFraction=-1;
	double scalefac=1.0;
	double lowpassNeighborWeight=.3;
	int status = 0; 
	
	/* 
		status==0 means leaksum is not present or else it is obsolete (don't write/read leaksum to/from disk cache)
	 	status==1 means leaksum is allocated and a deconvolver thread is currently accumulating leaked light into it
	 	status==2 means leaksum currently contains all the light that leaked into this section from others, and
	 				is ready to be subtracted, but has not yet been subtracted, from origSect;
	 	status==3 	means leaksum currently contains all the light that leaked into this section from others, and
	 				SubfacForSubtractionFraction*leaksum has been subtracted from origSect, after which
	 				origSect was rescaled by the factor scalefac. Thus with status==3,
	 				origSect = scalefac*(oldOrigSect - SubfacForSubtractionFraction*leaksum)
	 				Thus, to recover the prior version of origSect,
	 				oldOrigSect = (origSect/scalefac) + SubfacForSubtractionFraction*leaksum
	*/
	
	double mainParameter = .5; // for use only by getByteRaster()
	double totalLeakedLight=0;
	int ageSinceLastUsed=0;
	float [] bufline1=null,bufline2=null;
		
	public double [] computeErrorContributionCoeffs(float [] unprocessedData) {
		int i,brt,np,sum,num,avghigh,nhigh = (int)(.001*npix); // see the blurb at the beginning of econvolver.java to see what Aj and Bj are for
		double lsi,  upd,  CC=0.,  Aj =0,   Bj=0, gki, rampweight = .95/100.0;
		int [] hist = new int[1024];
		for(i=0;i<1024;i++) hist[i]=0;
		for(i=0;i<npix;i++) {
			lsi = leaksum[i];
			upd = unprocessedData[i];
			if((gki = 0.05 + rampweight*upd) >1.)
				gki=1.0; // the weight factor that emphasizes bright pixels
						// gki is never less than .05, and when upd==100, gki==1, its maximum
			Aj += gki*lsi*upd;
			Bj += gki*lsi*lsi;
			CC += gki*upd*upd;
			if((brt=(int)(.49+origSect[i])) > 1023) 
				brt = 1023;
			++hist[brt];
			}
		// find the brightest 1000 pixels 
		i=1023; np=1; sum=1;
		while(np<nhigh && i>1) {
			sum += hist[i]*i;
			np += hist[i];
			--i;
			}
		avghigh = sum/np; //this is the avg brightness of the brightest 0.1% of voxels in this section
		
		double [] ABC = new double[4];
		ABC[0]= Aj;
		ABC[1]= Bj;
		ABC[2]= CC;
		ABC[3] = 255.*(double)np/(double)sum;
		talk("For section "+secnum+", Aj="+Aj+", Bj="+Bj+", CC="+CC+", scaleFac = "+ABC[3]);
		return ABC;
		}		
		
	
	public synchronized void lowPassFilter(double nbhWt) {
		int i,row,col,rowoff;
		double sum;
		int [] offsets = {-raswidth-1, -raswidth, -raswidth+1,-1,1,raswidth-1,raswidth,raswidth+1};
		float [] buf,buff;// using these four arrays is just a way to avoid allocating and freeing
		// megabyte size arrays on every call to this method.
		int rwm = raswidth-1;
		boolean osc = false;
		lowpassNeighborWeight = nbhWt;
		if(bufline1==null) {
			bufline2 = new float[raswidth];
			bufline1 = new float[raswidth];
			}

		// preload bufline[] with first row
		for(i=0;i<raswidth;i++) 
			bufline1[i] = origSect[i];
		buf  = bufline2;
		buff = bufline1;			
			
		for(row=1;row<rasheight-1;row++) {
			rowoff=row*raswidth;
			if(osc) {
				buf  = bufline1;
				buff = bufline2;
				osc=false;			
				}
			else { // on first pass, buff=bufline1 which is what was preloaded
				buf  = bufline2;
				buff = bufline1;			
				osc=true;			
				}
			for(col=1;col<rwm;col++) {
				sum=0;
				for(i=0;i<8;i++)
					sum += origSect[rowoff+col+offsets[i]];
				buf[col] = (float)((origSect[rowoff+col] + nbhWt*sum)/(1.+8*nbhWt));
				}
			buf[0]=origSect[rowoff];
			buf[rwm]=origSect[rowoff+raswidth];
			// we'll never need the current row-1 again
			rowoff = (row-1)*raswidth;
			for(col=0;col<raswidth;col++) 
				origSect[rowoff+col] = buff[col];
			}
		// on exit from loop the last two rows of origSect remain unchanged
		rowoff = (rasheight-2)*raswidth;
		for(col=0;col<raswidth;col++) 
			origSect[rowoff+col] = buf[col];
		// the last row of origSect never gets changed. This method changes none of the border values of the raster.
		}
				
	
	public void lowPassFilterSimple(double nbhWt) {
		int i,row,col,rowoff;
		double sum;
		int [] offsets = {-raswidth-1, -raswidth, -raswidth+1,-1,1,raswidth-1,raswidth,raswidth+1};
		float [] bfr = new float[npix];
		for(row=1;row<rasheight-1;row++) {
			rowoff=row*raswidth;
			for(col=1;col<raswidth-1;col++) {
				sum=0;
				for(i=0;i<8;i++)
					sum += origSect[rowoff+col+offsets[i]];
				bfr[rowoff+col] = (float)((origSect[rowoff+col] + nbhWt*sum)/(1.+8*nbhWt));
				}
			}
		for(i=0;i<npix;i++)
			origSect[i] = bfr[i];
		bfr=null;
		}		
		
	public int age(boolean ifIncrement) {
		if(ifIncrement) 
			++ageSinceLastUsed;
		return (ageSinceLastUsed);
		}
	
	public void resetAge(int newage) {
		ageSinceLastUsed=newage;
		}		
		
	public int incrementIterationCounter() {
		++numDeconvolutionIterations;
		return numDeconvolutionIterations;
		}
	
	public void acceptLeaksum(float [] ls, int status) {
		if(leaksum==null) 
			leaksum = new float[npix];
		for(int i=0;i<npix;i++)
			leaksum[i] = ls[i];	
		this.status = status;
		}
					
	public String reportTotalLeakedLight() {
		if(status < 1) {
			talk("No leaksum[] array available. reportTotalLeakedLight() returns 0");
			return (" -777");
			}
		totalLeakedLight=0;
		for(int i=0;i<npix;i++)
			totalLeakedLight += leaksum[i];
		return(" "+trunc(totalLeakedLight)+" ");
		}
	
	public String trunc(double num) {
		// example num = -1023.1234567
		String stg = num <0.? "-" : ""; // stg = "-"
		num = Math.abs(num); // 1023.1234567
		int integralPart = (int)num; // 1023
		double remainder = num - integralPart; // 0.1234567
		int fourRemDigits = (int)(10000.*remainder); // 1234
		if(fourRemDigits==0) return (stg+integralPart+".0000");
		else if(fourRemDigits < 10)    return (stg+integralPart+".000"+fourRemDigits);
		else if(fourRemDigits < 100)   return (stg+integralPart+".00"+fourRemDigits);
		else if(fourRemDigits < 1000)  return (stg+integralPart+".0"+fourRemDigits);
		else return (stg+integralPart+"."+fourRemDigits);
		}
		
	public String statusReport() {
		double totalLeakedLight=0;
		String stg;
		if(status ==0) return("Section "+secnum+" has status = "+status+" and "+numDeconvolutionIterations+
			" deconvolution iterations. No leaked light info available");
		
		for(int i=0;i<npix;i++)
			totalLeakedLight += leaksum[i];
		stg = "sect "+secnum+" has status = "+status+" and "+numDeconvolutionIterations+" deconvolution iterations";
		if(status>0) 
			stg = stg + ", totalLeakedLight="+trunc(totalLeakedLight);
		return stg;
		}
				
	public LeakedLight(int raswidth, int rasheight, int secnum, int rgbchan, double mainParameter, float [] floatras) {
		this.raswidth = raswidth;
		this.rasheight = rasheight;
		this.secnum = secnum;
		this.rgbchan = rgbchan;
		npix = raswidth*rasheight;
		this.mainParameter =mainParameter;
		int pix;
		origSect = new float[npix];
		status=0;
		for(int i=0;i<npix;i++) 
			origSect[i]= floatras[i];
			
		numDeconvolutionIterations=0;
		resetAge(0);
		}
	
	public LeakedLight(int raswidth, int rasheight, int secnum, int rgbchan, double mainParameter) {
		// no creation of origSect, numDeconvolutionIterations left blank, used by LightSuckerClient()
		this.raswidth = raswidth;
		this.rasheight = rasheight;
		this.secnum = secnum;
		this.rgbchan = rgbchan;
		npix = raswidth*rasheight;
		this.mainParameter =mainParameter;
		resetAge(0);
		}
	
	public LeakedLight(String filename, double mainParameter) { // instantiate self from prior disk file
		readSectFromDisk(filename);
		this.mainParameter =mainParameter;
		resetAge(0);
		}
	
	public void sepuku() {
		leaksum=null;
		origSect=null;
		}
	
	private String trunc(float num) {
		// example num = -1023.1234567
		String stg = num <0.? "-" : ""; // stg = "-"
		num = Math.abs(num); // 1023.1234567
		int integralPart = (int)num; // 1023
		double remainder = num - integralPart; // 0.1234567
		int twoRemDigits = (int)(100.*remainder); // 12
		if(twoRemDigits==0) return (stg+integralPart+".00");
		else if(twoRemDigits < 10)    return (stg+integralPart+".0"+twoRemDigits);
		else return (stg+integralPart+"."+twoRemDigits);
		}
		
	public double totalLeakedLight() {
		int i;
		double sum=0;
		if(status<1) {
			talk("No leaksum[] array available. totalLeakedLight() returns 0");
			return 0.;
			}
		for(i=0;i<npix;i++)
			sum += leaksum[i];
		return sum;
		}
	
	public void zeroLeaksum() {
		int i;
		if(leaksum==null) 
			leaksum = new float[npix];
			
		for(i=0;i<npix;i++)
			leaksum[i]=0;
		status=1;
		SubfacForSubtractionFraction=-1;
		}
	
	public void scaleSection(double scalefac) {
		int i;
		this.scalefac = scalefac;
		for(i=0;i<npix;i++) 
				origSect[i] *= scalefac;
		}
		
	public byte [] getByteRaster(double boostFactor) {
		byte [] retn = new byte[npix];
		byte b255 = (byte)255;
		int scalebright;
		
		for(int i=0;i<npix; i++) {
			scalebright = (int)(.49+boostFactor*origSect[i]);
			if(scalebright>255)
				retn[i] = b255;
			else
				retn[i] = (byte)scalebright;
			}
		talk("LeakedLight.getByteRaster() return a "+retn.length+" byte raster");
		return retn;
		}	
			
	public byte [] getByteRaster(double GUIscaleFactor, byte [] rawDataSection) {
		byte [] retn = new byte[npix];
		byte b255 = (byte)255;
		int bright,i,np,sum,nhigh; 
		int [] hist = new int[1024];
		double boostFactor;
		
		// call this only when (status == 1 || status ==2) when leaked light not yet subtracted
		// so subtract it as best we can on the fly. 
		double subfrac = getSubfacForSubtractionFraction(mainParameter);
		for(i=0;i<1024;i++) 
			hist[i]=0;
		talk("generating partially deconvolved raster for section "+secnum+" using subfrac = "+subfrac);
		for(i=0;i<npix;i++) { //make the histogram of the data to be scaled.
			if((bright = (int)(origSect[i] - subfrac*leaksum[i])) <0) 
				bright = 0;
					
			else if(bright>1023)
					bright = 1023;
			++hist[bright];
			}

		// find the brightest 0.1% of pixels 
		i=1023; np=0; sum=0;
		nhigh = (int)(.001*npix);
		while(np<nhigh && i>1) {
			sum += hist[i]*i;
			np += hist[i];
			--i;
			}
		boostFactor = GUIscaleFactor*255.*(double)np/(double)sum;
// now scale the new iteration's deconvolved data estimate.
		for(i=0;i<npix;i++) { 
			if((bright = (int)(boostFactor*(origSect[i] - subfrac*leaksum[i]))) <0) 
				retn[i] = 0;
					
			else retn[i] = (bright > 255)? b255 : (byte)bright;
			}
		return retn;
		
		}	
		
	public byte [] getDoubleByteRaster(double GUIscaleFactor, byte [] rawDataSection) {
		byte [] retn = new byte[2*npix];
		byte b255 = (byte)255;
		int scalebright,higlow,loglow,j;
		int i,np,sum,bright,nhigh;
		int [] hist = new int[1024];
		double boostFactor;
		
		// call this only when (status == 1 || status ==2) when leaked light not yet subtracted
		// so subtract it as best we can on the fly. 
		talk("int LeakedLight.getDoubleByteRaster() putting low byte first");
		j=0;
		// so subtract it as best we can on the fly
		double subfrac = getSubfacForSubtractionFraction(mainParameter);
		for(i=0;i<1024;i++) 
			hist[i]=0;
		talk("generating partially deconvolved raster for section "+secnum+" using subfrac = "+subfrac);
		for(i=0;i<npix;i++) {
			if((bright = (int)(origSect[i] - subfrac*leaksum[i])) <0) 
				bright = 0;
			else if(bright>1023)
					bright = 1023;
			++hist[bright];
			}

		// find the brightest 0.1% of pixels 
		i=1023; np=0; sum=0;
		nhigh = (int)(.001*npix);
		while(np<nhigh && i>1) {
			sum += hist[i]*i;
			np += hist[i];
			--i;
			}
		boostFactor = GUIscaleFactor*256.*255.*(double)np/(double)sum;
// now scale the new iteration's deconvolved data estimate.
		j=0;
		for(i=0;i<npix;i++) { 
			if((bright = (int)(boostFactor*(origSect[i] - subfrac*leaksum[i]))) <0) 
				bright = 0;
					
			higlow = bright/256;
			loglow = bright - 256*higlow;
			retn[j] = (byte)loglow; // the low byte comes before the high byte
				++j;
				if(higlow >255)
					retn[j] = b255;
				else
					retn[j] = (byte)higlow;
				++j;
				}
		talk("LeakedLight.getDoubleByteRaster() returns a "+retn.length+" byte raster");
		return retn;
		}	
			
	public byte [] getDoubleByteRaster(double boostFactor) {
		byte [] retn = new byte[2*npix];
		byte b255 = (byte)255;
		int scalebright,higlow,loglow,j;
		double bright;
		talk("int LeakedLight.getDoubleByteRaster() putting low byte first");
		j=0;
		boostFactor = 256.*boostFactor;
		for(int i=0;i<npix; i++) {
			scalebright = (int)(.49 + boostFactor*origSect[i]);
			higlow = scalebright/256;
			loglow = scalebright - 256*higlow;
			retn[j] = (byte)loglow; // the low byte comes before the high byte
			++j;
			if(higlow >255)
				retn[j] = b255;
			else
				retn[j] = (byte)higlow;
			++j;
			}
		talk("LeakedLight.getDoubleByteRaster() returns a "+retn.length+" byte raster");
		return retn;
		}	
			
	public double getSubfacForSubtractionFraction(double frac) {
		int i,pix;
		double origtotal=0,leaktotal=0;
		if(status<1) {
			talk("No leaksum[] array available. getSubfacForSubtractionFraction() returns 0");
			return 0.;
			}
		for(i=0;i<npix;i++) {
			origtotal += origSect[i];
			leaktotal += leaksum[i];
			}
		// want subfac*leaktotal ~ frac*origtotal
		return frac*origtotal/(1.+leaktotal);
		}
			
	public double subtractLeakedLightFromSection(double subfac) {
		// returns percentage of unaccounted-for light because it wasn't there to be subtracted
		// CHANGES origSect[]
		if(status<1) {
			talk("No leaksum[] array available. subtractLeakedLightFromSection() returns 0");
			return 0.;
			}
		double darkmatter=0,origBrightMatter=0,bright;
		for(int i=0;i<npix;i++) {
			origBrightMatter += origSect[i];
			if((bright = origSect[i] - subfac*leaksum[i]) <0) {
					origSect[i]=0;
					darkmatter -= bright;
					}
				else 
					origSect[i] = (float)bright;
				}
		return darkmatter/(1+origBrightMatter);
		}	
	
	public double testSubtractLeakedLightFromSection(double subfac) {
		// returns percentage of unaccounted-for light because it wasn't there to be subtracted
		// does not change origSect[]
		if(status<1) {
			talk("No leaksum[] array available. testSubtractLeakedLightFromSection() returns 0");
			return 0.;
			}
		double darkmatter=0,origBrightMatter=0,bright;
		for(int i=0;i<npix;i++) {
			origBrightMatter += origSect[i];
			if((bright = origSect[i] - subfac*leaksum[i]) <0) 
					darkmatter -= bright;
			}
		return darkmatter/(1+origBrightMatter);
		}	

	public boolean writeSectToDisk(String filename) {
		talk("write self to disk file called:"+filename);
		// returns true if it worked
		FileOutputStream iostream;
		ObjectOutputStream p;
		try {
			iostream = new FileOutputStream(filename);
			p = new ObjectOutputStream(iostream);
			}
		catch(Exception e) {
			talk("Failed to open file <"+filename+"> got Exception = "+e);
			return false;
			}

		try {
			p.writeInt(raswidth);
			p.writeInt(rasheight);
			p.writeInt(secnum);
			p.writeInt(npix);
			p.writeInt(rgbchan);
			p.writeInt(numDeconvolutionIterations);
			p.writeInt(status);	
			p.writeDouble(SubfacForSubtractionFraction);		
			p.writeDouble(scalefac);
			p.writeDouble(lowpassNeighborWeight);		
			p.writeObject(origSect);
			p.writeObject(leaksum);
			}
		catch(Exception e) {
			talk("While writing section "+secnum+" to disk, got Exception = "+e);
			return false;
			}
		try {
			p.flush();
			iostream.close();
			}
		catch(Exception e) {
			talk("Got Exception = "+e);
			return false;
			}
		return true;
		}
					
	public boolean readSectFromDisk(String filename) {
		talk("read self from disk file called:"+filename);
		// returns true if it worked
		FileInputStream iostream;
		ObjectInputStream p;
		try {
			iostream = new FileInputStream(filename);
			p = new ObjectInputStream(iostream);
			}
		catch(Exception e) {
			talk("Failed to open file <"+filename+"> got Exception = "+e);
			return false;
			}

		try {
			raswidth = p.readInt();
			rasheight = p.readInt();
			secnum = p.readInt();
			npix = p.readInt();
			rgbchan = p.readInt();
			numDeconvolutionIterations = p.readInt();
			status =  p.readInt();
			SubfacForSubtractionFraction = p.readDouble();
			scalefac = p.readDouble();
			lowpassNeighborWeight = p.readDouble();
			origSect = (float[])p.readObject();
			if(status>0)
				leaksum = (float[])p.readObject();
			else leaksum=null;
			}
		catch(Exception e) {
			talk("While writing section "+secnum+" to disk, got Exception = "+e);
			return false;
			}
		try {
			iostream.close();
			}
		catch(Exception e) {
			talk("Got Exception = "+e);
			return false;
			}
		return true;
		}
					
	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
	}